'''
This example showcases RNG byte stream.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import rng_rava as rava

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    rava.lg.error('No device found')
    exit()

# Generate 3 bytes every 0.5s
rng.snd_rng_byte_stream_start(n_bytes=5, stream_interval_ms=500)

# Print 10 first values
print()
for i in range(10):
    rnd_a, rnd_b = rng.get_rng_byte_stream_data(list_output=True)
    print('RNG A, B = {}, {}'.format(rnd_a, rnd_b))

# Stop stream
rng.snd_rng_byte_stream_stop()

# Close device
rng.close()


'''
This example showcases the RNG generation of double precision floating-point
numbers.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import struct
import numpy as np
import rng_rava as rava

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    rava.lg.error('No device found')
    exit()

def get_rng_doubles(n_doubles):
    if n_doubles >= (2**32) // 8:
        print('RNG Doubles: Maximum n_doubles is {}'.format((2**32) // 8))
        return None

    # 64 bits floating point number
    bytes_res = rng.get_rng_bytes(n_bytes=n_doubles*8, timeout=None)
    if bytes_res is None:
        return
    rnd_bytes_a, rnd_bytes_b = bytes_res

    # XOR them
    int_a = int.from_bytes(rnd_bytes_a, 'little')
    int_b = int.from_bytes(rnd_bytes_b, 'little')
    rnd_bytes = (int_a ^ int_b).to_bytes(len(rnd_bytes_a), 'little')
    # Convert bytes to ints
    rnd_lists = struct.unpack('<{}Q'.format(n_doubles), rnd_bytes)
    rnd_ints = np.array(rnd_lists, dtype=np.uint64)

    # IEEE754 bit pattern for single precision floating point value in the
    # range of 1.0 - 2.0. Uses the first 52 bits and fixes the float
    # exponent to 1023
    rnd_ints_tmp = (rnd_ints & 0xFFFFFFFFFFFFF) | 0x3FF0000000000000
    rnd_bytes_filtered = rnd_ints_tmp.tobytes()
    rnd_lists_filtered = struct.unpack('<{}d'.format(n_doubles), rnd_bytes_filtered)
    rnd_doubles = np.array(rnd_lists_filtered, dtype=np.float64)
    return rnd_doubles - 1

# Get and print 100 doubles
doubles = get_rng_doubles(100)
doubles_str = '\n'.join(['{:.16f}'.format(d) for d in doubles])
print(doubles_str)

# Close device
rng.close()


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The RAVA_RNG_LED class implements the code for controlling the LED and LAMP
modules within the RAVA circuit. It allows users to adjust the color and the
intensity of the attached LED. Users can also activate the LAMP mode and
retrieve statistical information on its operation.

For more information on how LED and LAMP modules operate, refer to the firmware
documentation.
"""

from rng_rava.rava_defs import *
from rng_rava.rava_rng import *


### RAVA_RNG_LED

class RAVA_RNG_LED(RAVA_RNG):

    def __init__(self):
        super().__init__(dev_name='RAVA_LED')

        # Variables defined upon connection
        self.led_attached = None

        # Variables
        self.led_color = None
        self.led_intensity = None
        self.lamp_mode = None
        self.lamp_debug = None

        # Callback functions
        self.cbkfcn_lamp_debug = lambda: None


    def connect(self, serial_number):
        if not super().connect(serial_number=serial_number):
            return False

        # LED firmware enabled?
        if not self.led_enabled:
            lg.warning('{} Connect: LED code is disabled in the firmware'.format(self.dev_name))

        # LED attached?
        self.led_attached = self.get_eeprom_led()['led_attached']
        if not self.led_attached:
            lg.warning('{} Connect: Firmware claims no LED is attached.'
                       '\n{}  If false, fix it with snd_eeprom_led(led_attached=True)'
                       .format(self.dev_name, LOG_FILL))

        return True


    ## SERIAL

    def process_serial_comm(self, comm_id, comm_data):
        super().process_serial_comm(comm_id=comm_id, comm_data=comm_data)

        # LED_STATUS
        if comm_id == D_DEV_COMM['LED_STATUS']:
            # color, intensity, color_fading, intensity_fading
            self.put_queue_data('LED_STATUS', self.unpack_rava_msgdata(comm_data, 'BBBB'))

        # LAMP_DEBUG
        elif comm_id == D_DEV_COMM['LAMP_DEBUG']:
            # z, mag, mag_avg
            self.put_queue_data('LAMP_DEBUG', self.unpack_rava_msgdata(comm_data, 'fBB'))
            # Callback
            try:
                self.cbkfcn_lamp_debug()
            except:
                pass

        # LAMP_STATISTICS
        elif comm_id == D_DEV_COMM['LAMP_STATISTICS']:
            exp_n, exp_n_zsig, n_extra_bytes = self.unpack_rava_msgdata(comm_data, 'HHB')
            exp_colors_bytes = self.read_serial(n_extra_bytes)
            if exp_colors_bytes is None:     
                exp_colors = 8*[0]
            else:
                exp_colors = struct.unpack('<HHHHHHHH', exp_colors_bytes)
            self.put_queue_data('LAMP_STATISTICS', (exp_n, exp_n_zsig, *exp_colors))


    ## CALLBACK REGISTER

    def cbkreg_lamp_debug(self, fcn_lamp_debug):
        if callable(fcn_lamp_debug):
            self.cbkfcn_lamp_debug = fcn_lamp_debug
            lg.debug('{} Callback: Registering Lamp Debug function to {}'
                    .format(self.dev_name, fcn_lamp_debug.__name__))
        elif fcn_lamp_debug is None:
            self.cbkfcn_lamp_debug = lambda: None
            lg.debug('{} Callback: Unregistering Lamp Debug function'.format(self.dev_name))
        else:
            lg.error('{} Callback: Provide fcn_lamp_debug as a function'.format(self.dev_name))


    ## LED

    def snd_led_color(self, color_hue, intensity=255):
        if color_hue >= 2**8:
            lg.error('{} LED Color: Provide color_hue as a 8-bit integer'.format(self.dev_name))
            return None
        if intensity >= 2**8:
            lg.error('{} LED Color: Provide intensity as a 8-bit integer'.format(self.dev_name))
            return None

        self.led_color = color_hue
        self.led_intensity = intensity
        comm = 'LED_COLOR'
        return self.snd_rava_msg(comm, [color_hue, intensity], 'BB')


    def snd_led_color_fade(self, color_hue_tgt, duration_ms):
        if color_hue_tgt >= 2**8:
            lg.error('{} LED Color: Provide color_hue_tgt as a 8-bit integer'.format(self.dev_name))
            return None        
        if duration_ms >= 2**16:
            lg.error('{} LED Color: Provide duration_ms as a 16-bit integer'.format(self.dev_name))
            return None
        
        if duration_ms == 0:
            lg.error('{} LED Color: Provide duration_ms > 0'.format(self.dev_name))
            return None
        if color_hue_tgt - self.led_color == 0:
            return None

        self.led_color = color_hue_tgt
        comm = 'LED_COLOR_FADE'
        return self.snd_rava_msg(comm, [color_hue_tgt, duration_ms], 'BH')


    def snd_led_color_oscillate(self, n_cycles, duration_ms):
        if n_cycles >= 2**8:
            lg.error('{} LED Color: Provide n_cycles as a 8-bit integer'.format(self.dev_name))
            return None
        if duration_ms >= 2**16:
            lg.error('{} LED Color: Provide duration_ms as a 16-bit integer'.format(self.dev_name))
            return None
        
        if n_cycles == 0:
            lg.error('{} LED Color: Provide n_cycles > 0'.format(self.dev_name))
            return None
        if duration_ms == 0:
            lg.error('{} LED Color: Provide duration_ms > 0'.format(self.dev_name))
            return None

        comm = 'LED_COLOR_OSCILLATE'
        return self.snd_rava_msg(comm, [n_cycles, duration_ms], 'BH')


    def snd_led_intensity(self, intensity):
        if intensity >= 2**8:
            lg.error('{} LED Intensity: Provide intensity as a 8-bit integer'.format(self.dev_name))
            return None

        self.led_intensity = intensity
        comm = 'LED_INTENSITY'
        return self.snd_rava_msg(comm, [intensity], 'B')


    def snd_led_intensity_fade(self, intensity_tgt, duration_ms):
        if intensity_tgt >= 2**8:
            lg.error('{} LED Intensity: Provide intensity_tgt as a 8-bit integer'.format(self.dev_name))
            return None
        if duration_ms >= 2**16:
            lg.error('{} LED Intensity: Provide duration_ms as a 16-bit integer'.format(self.dev_name))
            return None
        
        if duration_ms == 0:
            lg.error('{} LED Intensity: Provide duration_ms > 0'.format(self.dev_name))
            return None
        if intensity_tgt - self.led_intensity == 0:
            return None

        self.led_intensity = intensity_tgt
        comm = 'LED_INTENSITY_FADE'
        return self.snd_rava_msg(comm, [intensity_tgt, duration_ms], 'BH')


    def snd_led_fade_stop(self):
        comm = 'LED_FADE_STOP'
        return self.snd_rava_msg(comm)


    def get_led_status(self, timeout=GET_TIMEOUT_S):
        comm = 'LED_STATUS'
        if self.snd_rava_msg(comm):
            color_hue, intensity, color_fading, intensity_fading = self.get_queue_data(comm, timeout=timeout)
            if color_hue in D_LED_COLOR_INV:
                color_str = D_LED_COLOR_INV[color_hue]
            else:
                color_str = ''
            return {'color_hue':color_hue, 'color':color_str, 'intensity':intensity,
                    'color_fading':color_fading, 'intensity_fading':intensity_fading}


    ## LAMP

    def snd_lamp_mode(self, on=True):
        self.lamp_mode = on

        comm = 'LAMP_MODE'
        return self.snd_rava_msg(comm, [on], 'B')


    def get_lamp_statistics(self, timeout=GET_TIMEOUT_S):
        # Every get command resets the statistics values
        comm = 'LAMP_STATISTICS'
        if self.snd_rava_msg(comm):
            data_vars = self.get_queue_data(comm, timeout=timeout)
            data_names = ['exp_n', 'exp_n_zsig',
                          'red', 'orange', 'yellow', 'green', 'cyan', 'blue', 'purple', 'pink']
            return dict(zip(data_names, data_vars))


    def snd_lamp_debug(self, on=True):
        self.lamp_debug = on
        comm = 'LAMP_DEBUG'
        return self.snd_rava_msg(comm, [on], 'B')


    def get_lamp_debug_data(self, timeout=GET_TIMEOUT_S):
        comm = 'LAMP_DEBUG'
        data_vars = self.get_queue_data(comm, timeout=timeout)

        # Timeout?
        if data_vars is None:
            return None

        data_names = ['z', 'mag', 'mag_avg']
        return dict(zip(data_names, data_vars))


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

from .rava_defs import *
from .rava_rng import *
from .rava_rng_aio import *
from .rava_rng_led import *


'''
This example showcases LED capabilities.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import time
import rng_rava as rava

# Find RAVA device and connect
rngled = rava.RAVA_RNG_LED()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rngled.connect(serial_number=dev_sns[0])
else:
    rava.lg.error('No device found')
    exit()

# Set color as blue (intensity=0)
print('> Blue color')
rngled.snd_led_color(color_hue=rava.D_LED_COLOR['BLUE'], intensity=0)

# Fade to full intensity
print('> Fade to full intensity')
rngled.snd_led_intensity_fade(intensity_tgt=255, duration_ms=2000)
time.sleep(3)

# Oscillate colors
print('> Oscillate colors')
rngled.snd_led_color_oscillate(n_cycles=3, duration_ms=4000)
time.sleep(5)

# Oscilate intensities
print('> Oscillate intensity')
for i in range(2):
    rngled.snd_led_intensity_fade(intensity_tgt=10, duration_ms=2000)
    time.sleep(2)
    rngled.snd_led_intensity_fade(intensity_tgt=255, duration_ms=2000)
    time.sleep(2)
time.sleep(1)

# Fade color to red
print('> Fade color to red')
rngled.snd_led_color_fade(color_hue_tgt=rava.D_LED_COLOR['RED'], duration_ms=2000)
time.sleep(3)

# Turn off lights
print('> Lights off')
rngled.snd_led_intensity(intensity=0)

# Close device
rngled.close()


'''
This example showcases the generation of a binary file containing random bytes.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import itertools
import rng_rava as rava

# Variables
FILE_OUTPUT = 'random.bin'
N_BYTES = 1000000 # 1MB
N_CHUNK = 10000

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    rava.lg.error('No device found')
    exit()

# Calculate n measurements
n_measurements = N_BYTES // N_CHUNK
n_bytes_remmaining = (N_BYTES % N_CHUNK)

# Open file
with open(FILE_OUTPUT, mode='bw') as f:

    # Loop over n measurements
    for i in range(n_measurements):
        print('{:.0f}%'.format(i / n_measurements * 100))

        # Generate bytes
        bytes_a, bytes_b = rng.get_rng_bytes(n_bytes=N_CHUNK // 2,
                                        postproc_id=rava.D_RNG_POSTPROC['NONE'],
                                        list_output=False,
                                        timeout=None)

        # Alternate RNG A and B bytes    
        bytes_ab = bytes(itertools.chain.from_iterable(zip(bytes_a, bytes_b)))
        
        # Write to file
        f.write(bytes_ab)

    # Remaining bytes
    if n_bytes_remmaining:
        
        # Generate bytes
        bytes_a, bytes_b = rng.get_rng_bytes(n_bytes=n_bytes_remmaining // 2,
                                        postproc_id=rava.D_RNG_POSTPROC['NONE'],
                                        list_output=False,
                                        timeout=None)

        # Alternate RNG A and B bytes    
        bytes_ab = bytes(itertools.chain.from_iterable(zip(bytes_a, bytes_b)))
        
        # Write to file
        f.write(bytes_ab)

    # Finished
    print('100%')

# Close device
rng.close()


'''
This example calculates the statistics of LAMP experiments over a specific time 
interval, yielding the rate of significant events over time. In the LAMP mode, 
each significant event is visually indicated by rapid color oscillations 
spanning the entire color spectrum.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import time
import rng_rava as rava

# Find RAVA device and connect
rngled = rava.RAVA_RNG_LED()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rngled.connect(serial_number=dev_sns[0])    
else:
    rava.lg.error('No device found')
    exit()


def measure_lamp_statistics():
    # Reset lamp stats
    if rngled.get_lamp_statistics() is None:
        return

    # Start lamp mode
    t0 = time.perf_counter()
    rngled.snd_lamp_mode(True)

    # Get elapsed time and lamp statistics
    input('\n> Press Enter to finish measuring lamp statistics...')
    t1 = time.perf_counter()
    stats = rngled.get_lamp_statistics()
    exp_n = stats['exp_n']

    # Close device
    rngled.close()

    # One experiment at least?
    if exp_n == 0:
        print('\nLamp Statistics: Not enougth LAMP experiments to process')
        return
    
    # Compute time intervals
    delta_t = t1 - t0
    delta_t_h = delta_t // 3600
    delta_t_min = (delta_t % 3600) // 60
    delta_t_s = delta_t % 60

    # Print results
    print('\nLamp Statistics' 
          '\nLasted {:.0f}h {:.0f}m {:.0f}s, yielding {} experiments'.format(delta_t_h, delta_t_min, delta_t_s, exp_n),
            
            '\n  where {} ({:.2f}%) reached statistical significance'.format(stats['exp_n_zsig'], 
                                                                             stats['exp_n_zsig']/exp_n*100),
                                  
            '\nMeaning one should expect to find:',
            
            '\n  {:.2f} significant events per 1 min'.format(stats['exp_n_zsig']/delta_t*60),
                    
            '\n  {:.2f} significant events per 5 min'.format(stats['exp_n_zsig']/delta_t*300),
            
            '\n  {:.2f} significant events per 10 min'.format(stats['exp_n_zsig']/delta_t*600),
            
            '\n  {:.2f} significant events per 30 min'.format(stats['exp_n_zsig']/delta_t*1800),
            
            '\n  {:.2f} significant events per 1 h'.format(stats['exp_n_zsig']/delta_t*3600),
            
            '\nColor distribution (%):',

            '\n  R={:.2f} O={:.2f} Y={:.2f} G={:.2f}'
            '\n  C={:.2f} B={:.2f} PU={:.2f} PI={:.2f}'
            .format(stats['red']/exp_n*100, stats['orange']/exp_n*100, stats['yellow']/exp_n*100, 
                    stats['green']/exp_n*100, stats['cyan']/exp_n*100, stats['blue']/exp_n*100, 
                    stats['purple']/exp_n*100, stats['pink']/exp_n*100))
    
measure_lamp_statistics()